"""
Downloads i2b2 data. Optionally also de-identifies it into a limited data set.
"""

from __future__ import annotations

import os
import logging
from logging import Logger
from pathlib import Path
from typing import TYPE_CHECKING, Union
if TYPE_CHECKING:
    from _typeshed.dbapi import DBAPIConnection
from typing_extensions import Literal
# Third-party packages
import pandas as pd
import pymssql
import sqlalchemy as sa
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)

# Arguments
COHORT_IDS_FILE_PATH = Path(r"..\..\..\work_2023_02_21\data\intermediate\createCohortFile_requestedCohort\2023-03-02 10-15-50\i2b2Cohort_fromRequestedCohort.CSV")
COHORT_ID_TYPE = "i2b2 Patient Number"  # A string. One of the `filderID` values of `getIDs`.
COHORT_COLUMN_NAME = "I2B2_PATIENT_NUM"  # A string. The column name of `COHORT_IDS_FILE_PATH` that contains the IDs.
COHORT_NAME = 'SGMCPLGB'  # A string. The name of cohort or study, e.g., "cancer_patients"
IRB_NUMBER = 'IRB201902162'  # A string. The IRB protocol and its number. E.g., "IRB123456789", "CED123456789", or "NH12345678". This is used in creating the patient de-identified IDs.
DE_IDENTIFY = None  # None-type or string "LDS", for "Limited data set".

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
USE_WINDOWS_AUTHENTICATION = True
SERVER = "EDW.shands.ufl.edu"
SERVER_I2B2 = "IDR01.shands.ufl.edu"
DATABASE = "DWS_PROD"
DATABASE_I2B2_GNV = "I2B2LTDDATA"
DATABASE_I2B2_JAX = "I2B2LTDDATAJAX"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"

# Variables: Map legacy variables to DRAPI-LEMUR standard variables
cohort = COHORT_NAME
irb = IRB_NUMBER

base_dir = projectDir
data_dir = dataDir
i2b2_dir = runOutputDir.joinpath("i2b2")
map_dir = runOutputDir.joinpath("mapping")
disclosure_dir = runOutputDir.joinpath("disclosure")

host = SERVER
database_prod = DATABASE
host_i2b2 = SERVER_I2B2
database_i2b2_GNV = DATABASE_I2B2_GNV
database_i2b2_JAX = DATABASE_I2B2_JAX

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Directory creation: Project-specific
makeDirPath(i2b2_dir)
if DE_IDENTIFY.lower():
    makeDirPath(map_dir)
    makeDirPath(disclosure_dir)
elif isinstance(DE_IDENTIFY, type(None)):
    pass
else:
    raise Exception("Invalid option for `DE_IDENTIFY`.")

# Functions: SQL


def connectToDatabase(host: str,
                      database: str,
                      useWindowsAuthentication=True) -> DBAPIConnection:
    """
    Connects to a SQL database given a `host` (server) and `database` value.
    """
    if useWindowsAuthentication:
        connection = pymssql.connect(host=host,
                                     database=database)
    else:
        conStr = f"mssql+pymssql://{uid}:{PWD}@{host}/{database}"
        connection = sa.create_engine(conStr).connect()
    return connection


def executeQuery(query: str, host: str, database: str) -> pd.DataFrame:
    """
    Executes a SQL query.
    INPUT:
        `query`: a string
        `host`: a string
        `databse`: a string
    OUTPUT:
        A pandas dataframe object.
    """
    databaseConnection = connectToDatabase(host, database)
    queryResult = pd.read_sql(query, databaseConnection)
    databaseConnection.close()
    return queryResult


# Functions: Script-specific
def getIDs(cohortFilePath: Union[Path, str],
           filterID: Literal["Patient Key", "EPIC Patient ID", "i2b2 Patient Number", "MRN (UF)", "MRN (Jax)"],
           logger: Logger,
           cohortColumnName: Union[None, str] = None) -> None:
    """
    Get i2b2 patient IDs from any of the following ID types:
        - Patient Key
        - EPIC Patient ID
        - i2b2 Patient Number
        - MRN (UF)
        - MRN (Jax)
    """

    if filterID == "Patient Key":
        filterStatement = "WHERE\n\tB.PATNT_KEY IN (XXXXX)"
        filterIDColumnName = filterID
    elif filterID == "EPIC Patient ID":
        filterStatement = "WHERE\n\tA.PATIENT_IDE IN (XXXXX)"
        filterIDColumnName = filterID
    elif filterID == "i2b2 Patient Number":
        filterStatement = "WHERE\n\tA.PATIENT_NUM IN (XXXXX)"
        filterIDColumnName = filterID
    elif filterID == "MRN (UF)":
        filterStatement = "WHERE\n\tC.IDENT_ID IN (XXXXX)"  # NOTE: Not tested
        filterIDColumnName = filterID
    elif filterID == "MRN (Jax)":
        filterStatement = "WHERE\n\tD.IDENT_ID IN (XXXXX)"  # NOTE: Not tested
        filterIDColumnName = filterID
    else:
        raise Exception("No valid option for `filterID` was passed.")

    if cohortColumnName:
        lookupColumnName = cohortColumnName
    else:
        lookupColumnName = filterIDColumnName

    query = f"""
    SELECT DISTINCT
        A.PATIENT_IDE AS 'EPIC Patient ID',
        A.PATIENT_NUM AS 'I2B2_PATIENT_NUM',
        B.PATNT_KEY AS 'Patient Key',
        C.IDENT_ID AS 'MRN (UF)',
        D.IDENT_ID AS 'MRN (Jax)'
    FROM
        [DWS_I2B2].[dbo].PATIENT_MAPPING A
        LEFT OUTER JOIN [DWS_PROD].[dbo].ALL_PATIENT_IDENTITIES B ON A.PATIENT_IDE = B.PATNT_ID
        LEFT OUTER JOIN [DWS_PROD].[dbo].ALL_PATIENT_IDENTITIES C ON A.PATIENT_IDE = C.PATNT_ID AND C.IDENT_ID_TYPE=101 AND C.LOOKUP_IND='Y'
        LEFT OUTER JOIN [DWS_PROD].[dbo].ALL_PATIENT_IDENTITIES D ON A.PATIENT_IDE = D.PATNT_ID AND D.IDENT_ID_TYPE=110 AND D.LOOKUP_IND='Y'
    {filterStatement}
    """
    m = 'w'
    h = True
    for chunk in pd.read_csv(COHORT_IDS_FILE_PATH, chunksize=3000):
        chunk = chunk[[lookupColumnName]]
        chunk = chunk.drop_duplicates()
        id = chunk[lookupColumnName].tolist()
        ids = "','".join(str(x) for x in id)
        ids = "'" + ids + "'"
        query1 = query.replace('XXXXX', ids)
        result = executeQuery(query1, host, database_prod)
        result = result.drop_duplicates()
        result.to_csv(os.path.join(data_dir, 'Cohort IDs.CSV'), index=False, mode=m, header=h)
        m = 'a'
        h = False
        logger.debug(query1)


def i2b2_dump_main(source, table, cohort_dir, i2b2_dir, cohort, logger):
    """
    source: 'GNV' or 'JAX'. Indicates which i2b2 instance to use as the source of data.
    table: 'patient_dimension', 'visit_dimension', or 'observation_fact'. Indicates which table to dump.
    cohort_dir: Specifies directory where cohort file is saved.
    i2b2_dir: directory where i2b2 data will be saved.
    cohort: The name of the cohort. E.g., 'subjects'.
    """
    h = True
    m = 'w'
    # Cohort should be saved in "Cohort IDs.CSV" file. It should contain i2b2 patient IDs in 'I2B2_PATIENT_NUM' column.
    for it, ids in enumerate(pd.read_csv(os.path.join(cohort_dir, 'Cohort IDs.CSV'), chunksize=100), start=1):
        logger.info(f"""  Working on batch {it:,}.""")
        ids = ids[['I2B2_PATIENT_NUM']].drop_duplicates().dropna()
        ids = ids['I2B2_PATIENT_NUM'].unique().tolist()
        ids = "','".join(str(x) for x in ids)
        ids = "'" + ids + "'"

        if (source == 'GNV'):
            database = 'I2B2LTDDATA'
        elif (source == 'JAX'):
            database = 'I2B2LTDDATAJAX'

        if (table == 'patient_dimension'):
            query = """
            select  PATIENT_NUM,VITAL_STATUS_CD,BIRTH_DATE,DEATH_DATE,SEX_CD,AGE_IN_YEARS_NUM,LANGUAGE_CD,RACE_CD,MARITAL_STATUS_CD,RELIGION_CD,ZIP_CD,STATECITYZIP_PATH,INCOME_CD,ETHNIC_CD,PAYER_CD,SMOKING_STATUS_CD,COUNTY_CD,SSN_VITAL_STATUS_CD,MYCHART_CD,CANCER_IND
            from database.dbo.PATIENT_DIMENSION
            where PATIENT_NUM in ( XXXXX )
            """
        elif (table == 'visit_dimension'):
            query = """
            select
            PATIENT_NUM,ENCOUNTER_NUM,ACTIVE_STATUS_CD,START_DATE,END_DATE,INOUT_CD,LOCATION_CD,LOCATION_PATH,LENGTH_OF_STAY
            from database.dbo.VISIT_DIMENSION
            where PATIENT_NUM in ( XXXXX )
            """
        elif (table == 'observation_fact'):
            query = """
            select  PATIENT_NUM,ENCOUNTER_NUM,CONCEPT_CD,START_DATE,MODIFIER_CD,VALTYPE_CD,TVAL_CHAR,NVAL_NUM,VALUEFLAG_CD,QUANTITY_NUM,UNITS_CD,END_DATE,LOCATION_CD
            from database.dbo.OBSERVATION_FACT
            where PATIENT_NUM in ( XXXXX )
            """
        query = query.replace('XXXXX', ids)
        query = query.replace('database', database)
        if (source == 'GNV'):
            result = executeQuery(query, host_i2b2, database_i2b2_GNV)
        elif (source == 'JAX'):
            result = executeQuery(query, host_i2b2, database_i2b2_JAX)
        result = result.drop_duplicates()
        file = cohort + '_' + table + '_' + source + '.csv'
        result.to_csv(os.path.join(i2b2_dir, file), mode=m, header=h, index=False)
        h = False
        m = 'a'
    logger.info(f"""Completed i2b2 dump for table "{table}" for location "{source}".""")


def generate_patient_map(map_dir, cohort_dir):
    pat = pd.read_csv(os.path.join(cohort_dir, 'Cohort IDs.CSV'))
    pat_map = pat[['Patient Key']].drop_duplicates()
    pat_map = pat_map.reset_index()
    pat_map['deid_num'] = pat_map.index + 1
    pat_map['deid_pat_ID'] = pat_map.apply(lambda row: str(irb) + '_PAT_' + str(int(row['deid_num'])), axis=1)
    pat = pd.merge(pat, pat_map, how='left', on='Patient Key')
    pat.to_csv(os.path.join(map_dir, 'map_patient.csv'), index=False)


def generate_encounter_map_i2b2(map_dir, i2b2_dir):
    # GNV
    in_file = cohort + '_visit_dimension_GNV.csv'
    df = pd.read_csv(os.path.join(i2b2_dir, in_file))
    df = df[['ENCOUNTER_NUM']].drop_duplicates()
    df1 = pd.DataFrame()
    in_file = cohort + '_observation_fact_GNV.csv'
    for chunk in pd.read_csv(os.path.join(i2b2_dir, in_file), chunksize=10000):
        chunk = chunk[['ENCOUNTER_NUM']].drop_duplicates()
        df1 = pd.concat([df1, chunk])
        df1 = df1.drop_duplicates()
    df = pd.concat([df, df1])
    df = df.drop_duplicates()
    df = df.reset_index()
    df['deid_num'] = df.index + 1
    df['deid_enc_ID'] = df.apply(lambda row: str(irb) + '_ENC_' + str(int(row['deid_num'])), axis=1)
    df['source'] = 'GNV'
    size = df.shape[0]

    # JAX
    in_file = cohort + '_visit_dimension_JAX.csv'
    df2 = pd.read_csv(os.path.join(i2b2_dir, in_file))
    df2 = df2[['ENCOUNTER_NUM']].drop_duplicates()
    df1 = pd.DataFrame()
    in_file = cohort + '_observation_fact_JAX.csv'
    for chunk in pd.read_csv(os.path.join(i2b2_dir, in_file), chunksize=10000):
        chunk = chunk[['ENCOUNTER_NUM']].drop_duplicates()
        df1 = pd.concat([df1, chunk])
        df1 = df1.drop_duplicates()
    df2 = pd.concat([df2, df1])
    df2 = df2.drop_duplicates()
    df2 = df2.reset_index()
    df2['deid_num'] = df2.index + 1 + size
    df2['deid_enc_ID'] = df2.apply(lambda row: str(irb) + '_ENC_' + str(int(row['deid_num'])), axis=1)
    df2['source'] = 'JAX'

    # Concatenate GNV and JAX
    df = pd.concat([df, df2])
    df.to_csv(os.path.join(map_dir, 'map_encounter.csv'), index=False)


def generate_mappings():
    generate_patient_map(map_dir, data_dir)
    generate_encounter_map_i2b2(map_dir, i2b2_dir)


def lds_i2b2_patient_dim(map_dir, i2b2_dir, disclosure_dir_i2b2, logger):
    map_pat = pd.read_csv(os.path.join(map_dir, 'map_patient.csv'))
    df = pd.DataFrame(columns=['deid_pat_ID', 'VITAL_STATUS_CD', 'BIRTH_DATE', 'DEATH_DATE', 'SEX_CD', 'AGE_IN_YEARS_NUM', 'LANGUAGE_CD', 'RACE_CD', 'MARITAL_STATUS_CD', 'RELIGION_CD', 'ZIP_CD', 'STATECITYZIP_PATH', 'INCOME_CD', 'ETHNIC_CD', 'PAYER_CD', 'SMOKING_STATUS_CD', 'COUNTY_CD', 'SSN_VITAL_STATUS_CD', 'MYCHART_CD', 'CANCER_IND'])
    df.to_csv(os.path.join(disclosure_dir_i2b2, 'patient_dimension.csv'), index=False)

    # GNV
    in_file = cohort + '_patient_dimension_GNV.csv'
    for df in pd.read_csv(os.path.join(i2b2_dir, in_file), chunksize=10000):
        df = df.drop_duplicates()
        df = pd.merge(df, map_pat, how='left', left_on='PATIENT_NUM', right_on='I2B2_PATIENT_NUM')
        df = df[['deid_pat_ID', 'VITAL_STATUS_CD', 'BIRTH_DATE', 'DEATH_DATE', 'SEX_CD', 'AGE_IN_YEARS_NUM', 'LANGUAGE_CD', 'RACE_CD', 'MARITAL_STATUS_CD', 'RELIGION_CD', 'ZIP_CD', 'STATECITYZIP_PATH', 'INCOME_CD', 'ETHNIC_CD', 'PAYER_CD', 'SMOKING_STATUS_CD', 'COUNTY_CD', 'SSN_VITAL_STATUS_CD', 'MYCHART_CD', 'CANCER_IND']]
        df.to_csv(os.path.join(disclosure_dir_i2b2, 'patient_dimension.csv'), header=False, index=False, mode='a')
    logger.info("De-identified GNV patient dimension.")

    # JAX
    in_file = cohort + '_patient_dimension_JAX.csv'
    for df in pd.read_csv(os.path.join(i2b2_dir, in_file), chunksize=10000):
        df = df.drop_duplicates()
        df = pd.merge(df, map_pat, how='left', left_on='PATIENT_NUM', right_on='I2B2_PATIENT_NUM')
        df = df[['deid_pat_ID', 'VITAL_STATUS_CD', 'BIRTH_DATE', 'DEATH_DATE', 'SEX_CD', 'AGE_IN_YEARS_NUM', 'LANGUAGE_CD', 'RACE_CD', 'MARITAL_STATUS_CD', 'RELIGION_CD', 'ZIP_CD', 'STATECITYZIP_PATH', 'INCOME_CD', 'ETHNIC_CD', 'PAYER_CD', 'SMOKING_STATUS_CD', 'COUNTY_CD', 'SSN_VITAL_STATUS_CD', 'MYCHART_CD', 'CANCER_IND']]
        df.to_csv(os.path.join(disclosure_dir_i2b2, 'patient_dimension.csv'), header=False, index=False, mode='a')
    logger.info("De-identified JAX patient dimension.")


def lds_i2b2_visit_dim(map_dir, i2b2_dir, disclosure_dir_i2b2, logger):
    map_pat = pd.read_csv(os.path.join(map_dir, 'map_patient.csv'))
    map_enc = pd.read_csv(os.path.join(map_dir, 'map_encounter.csv'))
    df = pd.DataFrame(columns=['deid_pat_ID', 'deid_enc_ID', 'ACTIVE_STATUS_CD', 'START_DATE', 'END_DATE', 'INOUT_CD', 'LOCATION_CD', 'LOCATION_PATH', 'LENGTH_OF_STAY'])
    df.to_csv(os.path.join(disclosure_dir_i2b2, 'visit_dimension.csv'), index=False)

    # GNV
    in_file = cohort + '_visit_dimension_GNV.csv'
    for df in pd.read_csv(os.path.join(i2b2_dir, in_file), chunksize=10000):
        df = df.drop_duplicates()
        df = pd.merge(df, map_pat, how='left', left_on='PATIENT_NUM', right_on='I2B2_PATIENT_NUM')
        df = pd.merge(df, map_enc, how='left', on='ENCOUNTER_NUM')
        df = df[['deid_pat_ID', 'deid_enc_ID', 'ACTIVE_STATUS_CD', 'START_DATE', 'END_DATE', 'INOUT_CD', 'LOCATION_CD', 'LOCATION_PATH', 'LENGTH_OF_STAY']]
        df.to_csv(os.path.join(disclosure_dir_i2b2, 'visit_dimension.csv'), header=False, index=False, mode='a')
    logger.info("De-identified GNV visit dimension.")

    # JAX
    in_file = cohort + '_visit_dimension_JAX.csv'
    for df in pd.read_csv(os.path.join(i2b2_dir, in_file), chunksize=10000):
        df = df.drop_duplicates()
        df = pd.merge(df, map_pat, how='left', left_on='PATIENT_NUM', right_on='I2B2_PATIENT_NUM')
        df = pd.merge(df, map_enc, how='left', on='ENCOUNTER_NUM')
        df = df[['deid_pat_ID', 'deid_enc_ID', 'ACTIVE_STATUS_CD', 'START_DATE', 'END_DATE', 'INOUT_CD', 'LOCATION_CD', 'LOCATION_PATH', 'LENGTH_OF_STAY']]
        df.to_csv(os.path.join(disclosure_dir_i2b2, 'visit_dimension.csv'), header=False, index=False, mode='a')
    logger.info("De-identified JAX visit dimension.")


def lds_i2b2_observation_fact(map_dir, i2b2_dir, disclosure_dir_i2b2, logger):
    map_pat = pd.read_csv(os.path.join(map_dir, 'map_patient.csv'))
    map_enc = pd.read_csv(os.path.join(map_dir, 'map_encounter.csv'))
    df = pd.DataFrame(columns=['deid_pat_ID', 'deid_enc_ID', 'CONCEPT_CD', 'START_DATE', 'MODIFIER_CD', 'VALTYPE_CD', 'TVAL_CHAR', 'NVAL_NUM', 'VALUEFLAG_CD', 'QUANTITY_NUM', 'UNITS_CD', 'END_DATE', 'LOCATION_CD'])
    df.to_csv(os.path.join(disclosure_dir_i2b2, 'observation_fact.csv'), index=False)

    CHUNK_SIZE = 10000

    # GNV
    in_file = cohort + '_observation_fact_GNV.csv'
    logger.info("""  ..  Reading file to count the number of chunks.""")
    numChunks = sum([1 for _ in pd.read_csv(in_file, chunksize=CHUNK_SIZE, dtype=str)])
    logger.info(f"""  ..  This file has {numChunks} chunks.""")
    fpath = os.path.join(i2b2_dir, in_file)
    dfChunks = pd.read_csv(fpath, chunksize=CHUNK_SIZE, low_memory=False)
    for it, df in enumerate(dfChunks, start=1):
        logger.info(f"""  Working on chunk {it:,} of {numChunks:,}.""")
        df = df.drop_duplicates()
        df = df[df['PATIENT_NUM'] != 'PATIENT_NUM']
        df['PATIENT_NUM'] = df['PATIENT_NUM'].astype(int)
        df = pd.merge(df, map_pat, how='left', left_on='PATIENT_NUM', right_on='I2B2_PATIENT_NUM')
        df = pd.merge(df, map_enc, how='left', on='ENCOUNTER_NUM')
        df = df[['deid_pat_ID', 'deid_enc_ID', 'CONCEPT_CD', 'START_DATE', 'MODIFIER_CD', 'VALTYPE_CD', 'TVAL_CHAR', 'NVAL_NUM', 'VALUEFLAG_CD', 'QUANTITY_NUM', 'UNITS_CD', 'END_DATE', 'LOCATION_CD']]
        df.to_csv(os.path.join(disclosure_dir_i2b2, 'observation_fact.csv'), header=False, index=False, mode='a')
    logger.info("De-identified GNV observation fact.")

    # JAX
    in_file = cohort + '_observation_fact_JAX.csv'
    logger.info("""  ..  Reading file to count the number of chunks.""")
    numChunks = sum([1 for _ in pd.read_csv(in_file, chunksize=CHUNK_SIZE, dtype=str)])
    logger.info(f"""  ..  This file has {numChunks} chunks.""")
    fpath = os.path.join(i2b2_dir, in_file)
    dfChunks = pd.read_csv(fpath, chunksize=CHUNK_SIZE, low_memory=False)
    for it, df in enumerate(dfChunks, start=1):
        logger.info(f"""  Working on chunk {it:,} of {numChunks:,}.""")
        df = df.drop_duplicates()
        df = df[df['PATIENT_NUM'] != 'PATIENT_NUM']
        df['PATIENT_NUM'] = df['PATIENT_NUM'].astype(int)
        df = pd.merge(df, map_pat, how='left', left_on='PATIENT_NUM', right_on='I2B2_PATIENT_NUM')
        df = pd.merge(df, map_enc, how='left', on='ENCOUNTER_NUM')
        df = df[['deid_pat_ID', 'deid_enc_ID', 'CONCEPT_CD', 'START_DATE', 'MODIFIER_CD', 'VALTYPE_CD', 'TVAL_CHAR', 'NVAL_NUM', 'VALUEFLAG_CD', 'QUANTITY_NUM', 'UNITS_CD', 'END_DATE', 'LOCATION_CD']]
        df.to_csv(os.path.join(disclosure_dir_i2b2, 'observation_fact.csv'), header=False, index=False, mode='a')
    logger.info("De-identified JAX observation fact.")


def lds_i2b2(map_dir, i2b2_dir, disclosure_dir_i2b2, logger):
    lds_i2b2_patient_dim(map_dir, i2b2_dir, disclosure_dir_i2b2, logger=logger)
    lds_i2b2_visit_dim(map_dir, i2b2_dir, disclosure_dir_i2b2, logger=logger)
    lds_i2b2_observation_fact(map_dir, i2b2_dir, disclosure_dir_i2b2, logger=logger)


def limited_data_set(logger):
    lds_i2b2(map_dir, i2b2_dir, disclosure_dir, logger=logger)


# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `COHORT_IDS_FILE_PATH`: "{COHORT_IDS_FILE_PATH}"
    `COHORT_NAME`: "{COHORT_NAME}"
    `IRB_NUMBER`: "{IRB_NUMBER}"

    # Arguments: SQL connection settings
    `USE_WINDOWS_AUTHENTICATION` : "{USE_WINDOWS_AUTHENTICATION}"
    `SERVER`                     : "{SERVER}"
    `SERVER_I2B2`                : "{SERVER_I2B2}"
    `DATABASE`                   : "{DATABASE}"
    `DATABASE_I2B2_GNV`          : "{DATABASE_I2B2_GNV}"
    `DATABASE_I2B2_JAX`          : "{DATABASE_I2B2_JAX}"
    `USERDOMAIN`                 : "{USERDOMAIN}"
    `USERNAME`                   : "{USERNAME}"
    `UID`                        : "{UID}"
    `PWD`                        : censored

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    """)

    # Generate i2b2 patient IDs
    logger.info("Generating i2b2 patient IDs.")
    getIDs(cohortFilePath=COHORT_IDS_FILE_PATH,
           cohortColumnName=COHORT_COLUMN_NAME,
           filterID=COHORT_ID_TYPE,
           logger=logger)
    logger.info("Generating i2b2 patient IDs - done.")

    # Perform i2b2 dump
    makeDirPath(i2b2_dir)
    i2b2_dump_main('GNV', 'patient_dimension', data_dir, i2b2_dir, cohort, logger=logger)  # Pull data from patient_dimension in GNV i2b2 instance.
    i2b2_dump_main('JAX', 'patient_dimension', data_dir, i2b2_dir, cohort, logger=logger)
    i2b2_dump_main('GNV', 'visit_dimension', data_dir, i2b2_dir, cohort, logger=logger)
    i2b2_dump_main('JAX', 'visit_dimension', data_dir, i2b2_dir, cohort, logger=logger)
    i2b2_dump_main('GNV', 'observation_fact', data_dir, i2b2_dir, cohort, logger=logger)
    i2b2_dump_main('JAX', 'observation_fact', data_dir, i2b2_dir, cohort, logger=logger)

    # Prepare limited data set for disclosure
    if DE_IDENTIFY.lower() == "lds":
        generate_mappings()
        limited_data_set(logger=logger)
    elif isinstance(DE_IDENTIFY, type(None)):
        pass
    else:
        raise Exception("Invalid option for `DE_IDENTIFY`.")

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(projectDir)}".""")



"""
Searches and selects notes by the list of note types contained in `NOTE_TYPES_TO_FILTER_BY` and outputs the corresponding metadata and notes text.

Note we distinguish between the following three concepts
    - Note Class (two values)
    - Note Category (four values)
    - Note Type (More than 100 values for each class.)
See the NOTE comment tags in the code for details.
"""

import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)

# Arguments
FREE_TEXT_DIR_PATH = Path(r"data\output\freeText\...\free_text")  # TODO
COHORT_NAME = ""  # TODO
NOTE_TYPES_TO_FILTER_BY = ["order_impression: IMAGING",
                           "order_narative: IMAGING"]  # TODO
DE_IDENTIFICATION_MAP_PATH_FOR_NOTES = r"data\output\freeText\...\mapping\map_note_link.csv"  # TODO
DE_IDENTIFICATION_MAP_PATH_FOR_ORDERS = r"data\output\freeText\...\mapping\map_order.csv"  # TODO
USE_DE_IDENTIFIED_NOTE_VERSION = True  # TODO
CHUNKSIZE = 10000


# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
# ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
#  "IRB_DIRECTORY",                 # noqa
#  "DATA_REQUEST_DIRECTORY",        # noqa
#  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
noteDeIdentificationMaps = {"Notes": DE_IDENTIFICATION_MAP_PATH_FOR_NOTES,
                            "Orders": DE_IDENTIFICATION_MAP_PATH_FOR_ORDERS}

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `FREE_TEXT_DIR_PATH`: "{FREE_TEXT_DIR_PATH}"
    `COHORT_NAME`: "{COHORT_NAME}"
    `NOTE_TYPES_TO_FILTER_BY`: "{NOTE_TYPES_TO_FILTER_BY}"
    `DE_IDENTIFICATION_MAP_PATH_FOR_NOTES`: "{DE_IDENTIFICATION_MAP_PATH_FOR_NOTES}"
    `DE_IDENTIFICATION_MAP_PATH_FOR_ORDERS`: "{DE_IDENTIFICATION_MAP_PATH_FOR_ORDERS}"
    `USE_DE_IDENTIFIED_NOTE_VERSION`: "{USE_DE_IDENTIFIED_NOTE_VERSION}"
    `CHUNKSIZE`: "{CHUNKSIZE}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Search by note types
    logger.info("Searching by note types.")
    LIST_OF_NOTE_TARGET_FILES_AND_DIRS = ["_note"]
    LIST_OF_ORDER_TARGET_FILES_AND_DIRS = ["_order_impression",
                                           "_order",
                                           "_order_narrative",
                                           "_order_result_comment"]
    LIST_OF_TARGET_FILES_AND_FOLDERS = LIST_OF_NOTE_TARGET_FILES_AND_DIRS + LIST_OF_ORDER_TARGET_FILES_AND_DIRS
    resultsdi = {}
    for fpath in FREE_TEXT_DIR_PATH.iterdir():
        logger.info(f"""  Working on directory item "{fpath.absolute().relative_to(rootDirectory)}".""")
        if fpath.suffix.lower() == ".csv" and any([fname in fpath.stem for fname in LIST_OF_TARGET_FILES_AND_FOLDERS]):
            logger.info("""    This item has met the criteria for processing.""")
            numChunks = 0
            logger.info("""    Counting the number of chunks in the file.""")
            for dfchunk in pd.read_csv(fpath, chunksize=CHUNKSIZE):
                numChunks += 1
            logger.info("""    Counting the number of chunks in the file - done.""")
            fileResults = pd.DataFrame()
            for it, dfchunk in enumerate(pd.read_csv(fpath, chunksize=CHUNKSIZE), start=1):
                logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                mask = dfchunk["NoteType"].isin(NOTE_TYPES_TO_FILTER_BY)
                subset = dfchunk[mask]
                fileResults = pd.concat([fileResults, subset])
            resultsdi[fpath.stem] = fileResults.drop_duplicates()
        else:
            logger.info("""    This item has NOT met the criteria for processing.""")

    # Save search results
    logger.info("Saving search results to the drive.")
    for noteCategory, df in resultsdi.items():
        rsavepath = runOutputDir.joinpath("Selected Notes Metadata",
                                          f"{noteCategory}.CSV")
        makeDirPath(rsavepath.parent)
        df.to_csv(rsavepath, index=False)
    logger.info("Saving search results to the drive - done.")

    # Concatenate results by note class
    logger.info("Concatenating results by note class.")
    results = {"Notes": None,
               "Orders": None}
    # NOTE: Here we introduce the "Note Category" concept, which is one of four values: "Note", "Order Impression", "Order Narrative" or "Order Result Comment".
    for noteCategory, df in resultsdi.items():
        if any([name in noteCategory for name in LIST_OF_NOTE_TARGET_FILES_AND_DIRS]):
            results["Notes"] = pd.concat([results["Notes"], df["LinkageNoteID"]])
        elif any([name in noteCategory for name in LIST_OF_ORDER_TARGET_FILES_AND_DIRS]):
            results["Orders"] = pd.concat([results["Orders"], df["OrderKey"]])
        else:
            raise Exception("Something unexpected happened.")
    # NOTE: Here we introduce the "Note Class" concept, which is one of two values: "Note" or "Order".
    for noteClass, series in results.items():
        series = pd.Series(series)
        results[noteClass] = series.drop_duplicates().sort_values()
    logger.info("Concatenating results by note class - done.")

    # For each note class, filter by its corresponding note identifier, i.e., `LinkageNoteID` or `OrderKey`
    for noteClass, noteIdentifierValues in results.items():
        logger.info(f"""  Working on note class "{noteClass}".""")
        # Get note class identifier map
        if noteClass == "Notes":
            NOTE_IDENTIFIER_TYPE = "LinkageNoteID"
            NOTE_IDENTIFIER_TYPE_DE_IDENTIFIED = "deid_link_note_id"
        elif noteClass == "Orders":
            NOTE_IDENTIFIER_TYPE = "OrderKey"
            NOTE_IDENTIFIER_TYPE_DE_IDENTIFIED = "deid_order_id"
        noteIdentifierValues = pd.DataFrame(noteIdentifierValues)
        noteIdentifierValues = noteIdentifierValues.set_index(NOTE_IDENTIFIER_TYPE).sort_index()
        mapPath = noteDeIdentificationMaps[noteClass]
        noteIdentifierMap = pd.read_csv(mapPath)
        noteIdentifierMap = noteIdentifierMap.set_index(NOTE_IDENTIFIER_TYPE)
        selectedNoteIdentiferMap = noteIdentifierValues.join(noteIdentifierMap, how="inner")
        selectedNoteIdentiferMap = selectedNoteIdentiferMap.reset_index()
        selectedNoteIdentiferMap = selectedNoteIdentiferMap.set_index(NOTE_IDENTIFIER_TYPE_DE_IDENTIFIED)

        # Join map on note text files.
        DICT_OF_TARGET_FILES_AND_FOLDERS = {"Notes": LIST_OF_NOTE_TARGET_FILES_AND_DIRS,
                                            "Orders": LIST_OF_ORDER_TARGET_FILES_AND_DIRS}
        if USE_DE_IDENTIFIED_NOTE_VERSION:
            noteDeIdentificationPrefix = "deid"
        else:
            noteDeIdentificationPrefix = ""
        for fpath in FREE_TEXT_DIR_PATH.glob("**/*"):
            logger.info(f"""    Working on directory item "{fpath.absolute().relative_to(rootDirectory)}".""")
            listOfClassTargetFilesAndFolders = DICT_OF_TARGET_FILES_AND_FOLDERS[noteClass]
            if fpath.suffix.lower() == ".tsv" and any([fname in fpath.stem for fname in listOfClassTargetFilesAndFolders]) and "deid_" in fpath.stem:
                logger.info("""  ..  This item has met the criteria for processing.""")
                # Determine note identifier
                if any([f"{noteDeIdentificationPrefix}{nameRoot}" in fpath.stem for nameRoot in LIST_OF_NOTE_TARGET_FILES_AND_DIRS]):
                    noteIdentifierDeIdentifiedColumnName = "deid_link_note_id"
                elif any([f"{noteDeIdentificationPrefix}{nameRoot}" in fpath.stem for nameRoot in LIST_OF_ORDER_TARGET_FILES_AND_DIRS]):
                    noteIdentifierDeIdentifiedColumnName = "deid_order_id"
                else:
                    raise Exception("Something went wrong.")
                # Count number of chunks
                logger.info("""  ..  Counting the number of chunks in the file.""")
                numChunks = 0
                for dfchunk in pd.read_csv(fpath, chunksize=CHUNKSIZE, delimiter="\t"):
                    numChunks += 1
                # Iterate over chunks
                logger.info("""  ..  Counting the number of chunks in the file - done.""")
                header = True
                mode = "w"
                for it, dfchunk in enumerate(pd.read_csv(fpath, chunksize=CHUNKSIZE, delimiter="\t"), start=1):
                    logger.info(f"""  ..    Working on chunk {it} of {numChunks}.""")
                    # Filter by selected note or order identifiers.
                    dfchunk = dfchunk.set_index(noteIdentifierDeIdentifiedColumnName)
                    selectedText = selectedNoteIdentiferMap.join(dfchunk, how="inner")
                    selectedText = selectedText["note_text"]
                    # Define `savepath` based on `fpath`
                    savepath = runOutputDir.joinpath("Selected Notes",
                                                     "{fpath.stem}.TSV")
                    makeDirPath(savepath.parent)
                    # Save to file
                    selectedText = selectedText.reset_index()
                    selectedText.to_csv(savepath, index=False, header=header, mode=mode, sep="\t")
                    header = False
                    mode = "a"
                logger.info(f"""  ..  Chunks saved to "{savepath.absolute().relative_to(rootDirectory)}".""")
            else:
                logger.info("""  ..  This item has NOT met the criteria for processing.""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.relative_to(projectDir)}".""")



"""
Herman's utility functions commonly used in his IDEALIST projects
"""

__all__ = ["idealistMap2dict",
           "patientKey2MRNs"]

from pathlib import Path
from drapi.code.drapi.drapi import replace_sql_query
import os
import pandas as pd
import sqlalchemy as sa

# Variables
this_file_path = Path(__file__)
this_file_stem = this_file_path.stem
project_dir = this_file_path.absolute().parent.parent.parent.parent
input_dir = project_dir.joinpath("data", "input")
output_dir = project_dir.joinpath("data", "output")
sql_dir = project_dir.joinpath("sql")


# SQL Server settings
SERVER = "DWSRSRCH01.shands.ufl.edu"  # AKA `HOST`
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = fr"{USERDOMAIN}\{USERNAME}"
PWD = os.environ["HFA_UFADPWD"]

# SQLAlchemy connections
connstr = f"mssql+pymssql://{UID}:{PWD}@{SERVER}/{DATABASE}"  # Create connection string
engine = sa.create_engine(connstr)  # Make connection/engine


def idealistMap2dict(idealistMap, fromID, toID):
    """
    """
    idealistMap.index = idealistMap[fromID]
    mapDi = idealistMap[toID].to_dict()
    return mapDi


def patientKey2MRNs(patientKeys: list):
    """
    """
    assert isinstance(patientKeys, list), "This function expects an object of type list."
    queryPath = sql_dir.joinpath("MRNfromPatientKey.sql")
    with open(queryPath, "r") as file:
        query0 = file.read()
    old = "0123456789, 1234567890"
    new = ",".join(str(id) for id in patientKeys)
    query = replace_sql_query(query0, old, new)
    results = pd.read_sql(query, engine)
    return results



"""
Create a map from OneFlorida Patient IDs to IDR Patient Keys
"""

import argparse
import logging
import os
import pprint
from pathlib import Path
# Third-party packages
import pandas as pd
from sqlalchemy import URL
# Local packages
from drapi import __version__ as drapiVersion
from drapi import PATH as drapiInstallationPath
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    makeDirPath,
                                    replace_sql_query)
from drapi.code.drapi.getData.getData import getData
from drapi.code.drapi.oneFlorida import ID_TYPE_DICT as ONE_FLORIDA_ID_TYPE_DICT
from drapi.code.drapi.oneFlorida import ID_TYPE_LIST as ONE_FLORIDA_ID_TYPE_LIST


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("--FILE_PATH",
                        help="The path to the file that is used as input. The file can contain either OneFlorida patient IDs or IDR patient IDs.")
    parser.add_argument("--FILE_HEADER",
                        help="The header of the file containing the variable you want to convert from.")
    parser.add_argument("--FROM",
                        type=str,
                        choices=ONE_FLORIDA_ID_TYPE_LIST,
                        help="The variable used as input.")
    parser.add_argument("--TO",
                        type=str,
                        choices=ONE_FLORIDA_ID_TYPE_LIST,
                        help="The variable to be output.")
    parser.add_argument("--ID_TYPE",
                        type=str,
                        choices=ONE_FLORIDA_ID_TYPE_LIST,
                        help="The standard variable name use to filter the SQL query.")
    parser.add_argument("--FIRST_TIME",
                        action="store_true",
                        help="""If you have run this script before and have downloaded the data, select "True", otherwise select "False".""")
    parser.add_argument("--OLD_RUN_PATH",
                        help="The path to the directory that contains the downloaded intermediate data.")

    # Arguments: Main
    parser.add_argument("--SCRIPT_TEST_MODE",
                        action="store_true",
                        help="""Use this option to run a shorter version.""")

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-parameters
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    FILE_PATH = argNamespace.FILE_PATH
    FILE_HEADER = argNamespace.FILE_HEADER
    FROM = argNamespace.FROM
    TO = argNamespace.TO
    ID_TYPE = argNamespace.ID_TYPE
    FIRST_TIME = argNamespace.FIRST_TIME
    OLD_RUN_PATH = argNamespace.OLD_RUN_PATH

    # >>> meta variable <<<
    FORM_1_ARGUMENTS = [FIRST_TIME,
                        FILE_PATH,
                        FILE_HEADER,
                        FROM,
                        TO,
                        ID_TYPE]
    FORM_2_ARGUMENTS = [OLD_RUN_PATH,
                        FROM,
                        TO,
                        ID_TYPE]
    # <<< meta variable <<<

    SCRIPT_TEST_MODE = argNamespace.SCRIPT_TEST_MODE

    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-parameters
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    if FIRST_TIME and OLD_RUN_PATH:
        message = "It is ambiguous if you provide both `FIRST_TIME` and `OLD_RUN_PATH`. Please only choose one."
        parser.error(message=message)
    # <<< Argument checks <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        userID = USER_ID[:]
    else:
        userID = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPwd = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{userID}:{userPwd}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin module body >>>

    # Download OneFlorida to IDR map
    if FIRST_TIME:
        # Select input: NOTE This block is structured like an else-if block to remind the programmer of how he can use `getData`.
        # Select input: Dataframe
        if False:
            fromData = pd.read_csv(filepath_or_buffer=FILE_PATH)
            filePath = None
            fileHeader = None
        # Select input: Series
        elif False:
            fromData = pd.read_csv(filepath_or_buffer=FILE_PATH)[FILE_HEADER]
            filePath = None
            fileHeader = None
        # Select input: File path with column name
        elif True:
            fromData = None
            filePath = FILE_PATH
            fileHeader = FILE_HEADER
        else:
            raise Exception("This should not happen")

        downloadDir = runIntermediateDir.joinpath("Downloaded Data")
        downloadDir.mkdir()

        connectionString = URL.create(drivername="mssql+pymssql",
                                      username=userID,
                                      password=userPwd,
                                      host=SERVER,
                                      database=DATABASE)
        
        # Modify query according to `TO` and `FROM` arguments
        SQL_FILE_PATH_RELATIVE = Path("../../src/drapi/sql/OneFlorida to UF Health Patient ID Map.SQL")
        SQL_FILE_PATH = drapiInstallationPath.joinpath(SQL_FILE_PATH_RELATIVE)
        with open(SQL_FILE_PATH, "r") as file:
            query0 = file.read()
            IDTypeInput = ID_TYPE.lower()
            IDTypeSQL = ONE_FLORIDA_ID_TYPE_DICT[IDTypeInput]
            query = replace_sql_query(query=query0,
                                      old="{PYTHON_VARIABLE: IDTypeSQL}",
                                      new=IDTypeSQL,
                                      logger=logger)
        sqlFilePathTemp = runIntermediateDir.joinpath(f"OneFlorida to UF Health Patient ID Map - {FROM}.SQL")
        with open(sqlFilePathTemp, "w") as file:
            file.write(query)

        getData(sqlFilePath=sqlFilePathTemp,
                connectionString=connectionString,
                filterVariableChunkSize=10000,
                filterVariableColumnName=fileHeader,
                filterVariableData=fromData,
                filterVariableFilePath=filePath,
                filterVariablePythonDataType="int",
                filterVariableSqlQueryTemplatePlaceholder="{PYTHON_VARIABLE: IDTypeValues}",
                logger=logger,
                outputFileName=f"OneFlorida to UF Health Patient ID Map - {FROM}",
                runOutputDir=downloadDir,
                queryChunkSize=10000)
    else:
        downloadDir = Path(OLD_RUN_PATH)

    # Concatenate downloaded data
    pathli = sorted(list(downloadDir.iterdir()))
    concatenatedMapPath = runOutputDir.joinpath(f"OneFlorida to UF Health Patient ID Map - {FROM}.CSV")
    header = True
    for fpath in pathli:
        logger.info(f"""  Reading file "{choosePathToLog(path=fpath, rootPath=projectDir)}".""")
        df = pd.read_csv(filepath_or_buffer=fpath)
        df.to_csv(path_or_buf=concatenatedMapPath,
                  index=False,
                  header=header,
                  mode="a")
        header = False

    # Select data to output
    columnsToExport = [TO,
                       FROM]
    finalMap = pd.read_csv(filepath_or_buffer=concatenatedMapPath)
    finalMap = finalMap[columnsToExport]
    finalMap = finalMap.sort_values(by=TO)
    exportPath = runOutputDir.joinpath(f"Map - {FROM} to {TO}.CSV")
    finalMap.to_csv(exportPath, index=False)

    # QA
    mapSize = finalMap.dropna().shape[0]
    logger.info(f"""Final map shape after dropping any NAs: {mapSize:,}.""")

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # <<< End module body <<<
    logger.info(f"""Finished running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")



"""

A template for pulling SQL queries in Python

Outline:

  - Load arguments
  - Load query
  - Insert arguments into query
  - Run query
  - Post-processing
"""

import os
import pandas as pd
import sqlalchemy as sa
from drapi.drapi import getTimestamp, isValidPatientID, replace_sql_query
from pathlib import Path

# Arguments
pass

# Variables
this_file_path = Path(__file__)
project_dir = this_file_path.absolute().parent.parent
irb_dir = project_dir.parent
irbNumber = None  # TODO
input_dir = os.path.join(project_dir, "data",
                                      "input")
output_dir = os.path.join(project_dir, "data",
                                       "output")
sql_dir = os.path.join(project_dir, "sql")
run_timestamp = getTimestamp()

# SQL Server settings
SERVER = "DWSRSRCH01.shands.ufl.edu"  # AKA `HOST`
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = fr"{USERDOMAIN}\{USERNAME}"
PWD = os.environ["HFA_UFADPWD"]

# SQLAlchemy connection
connstr = f"mssql+pymssql://{UID}:{PWD}@{SERVER}/{DATABASE}"  # Create connection string
engine = sa.create_engine(connstr)  # Make connection/engine

# Test connections
if True:
    # SQL
    print(f"""[{getTimestamp()}] Testing SQL connection.""")
    query_test = """SELECT 1"""
    test = pd.read_sql(query_test, con=engine).values[0][0]
    if test:
        print(f"""[{getTimestamp()}] SQL connection successful: "{test}".""")



"""
Get all variables/columns of tables/files in the project.
"""

import logging
import os
import sys
from collections import OrderedDict
from pathlib import Path
# Third-party packages
import numpy as np
import pandas as pd
from colorama import init as colorama_init
from colorama import (Fore,
                      Style)
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
# Project parameters:
from common import DATA_REQUEST_ROOT_DIRECTORY_DEPTH
# Project parameters: Portion paths and criteria
from common import (BO_PORTION_DIR_MAC, BO_PORTION_DIR_WIN,
                    I2B2_PORTION_DIR_MAC, I2B2_PORTION_DIR_WIN,
                    MODIFIED_OMOP_PORTION_DIR_MAC, MODIFIED_OMOP_PORTION_DIR_WIN,
                    NOTES_PORTION_DIR_MAC, NOTES_PORTION_DIR_WIN,
                    OMOP_PORTION_DIR_MAC, OMOP_PORTION_DIR_WIN)
# Project parameters: Criteria
from common import (BO_PORTION_FILE_CRITERIA,
                    I2B2_PORTION_FILE_CRITERIA,
                    NOTES_PORTION_FILE_CRITERIA,
                    OMOP_PORTION_FILE_CRITERIA)
from drapi.code.drapi.constants.phiVariables import (LIST_OF_PHI_DATES_BO,
                                          LIST_OF_PHI_DATES_NOTES,
                                          LIST_OF_PHI_DATES_OMOP)

# Arguments: OMOP data set selection
USE_MODIFIED_OMOP_DATA_SET = True

# Arguments: Portion Paths and conditions
if USE_MODIFIED_OMOP_DATA_SET:
    OMOPPortionDirMac = MODIFIED_OMOP_PORTION_DIR_MAC
    OMOPPortionDirWin = MODIFIED_OMOP_PORTION_DIR_WIN
else:
    OMOPPortionDirMac = OMOP_PORTION_DIR_MAC
    OMOPPortionDirWin = OMOP_PORTION_DIR_WIN

PORTIONS_OUTPUT_DIR_PATH_MAC = {"BO": BO_PORTION_DIR_MAC,  # TODO
                                "i2b2": I2B2_PORTION_DIR_MAC,
                                "Clinical Text": NOTES_PORTION_DIR_MAC,
                                "OMOP": OMOP_PORTION_DIR_MAC}
PORTIONS_OUTPUT_DIR_PATH_WIN = {"BO": BO_PORTION_DIR_WIN,  # TODO
                                "i2b2": I2B2_PORTION_DIR_WIN,
                                "Clinical Text": NOTES_PORTION_DIR_WIN,
                                "OMOP": OMOP_PORTION_DIR_WIN}
PORTION_FILE_CRITERIA_DICT = {"BO": BO_PORTION_FILE_CRITERIA,
                              "i2b2": I2B2_PORTION_FILE_CRITERIA,
                              "Clinical Text": NOTES_PORTION_FILE_CRITERIA,
                              "OMOP": OMOP_PORTION_FILE_CRITERIA}

COMPARISON_SET = LIST_OF_PHI_DATES_BO + LIST_OF_PHI_DATES_NOTES + LIST_OF_PHI_DATES_OMOP

# Arguments: Meta-variables
LOG_LEVEL = "DEBUG"

CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = np.all([path.exists() for path in PORTIONS_OUTPUT_DIR_PATH_MAC.values()]) or np.all([path.exists() for path in PORTIONS_OUTPUT_DIR_PATH_WIN.values()])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        portionsOutputDirPath = PORTIONS_OUTPUT_DIR_PATH_MAC
    elif operatingSystem == "win32":
        portionsOutputDirPath = PORTIONS_OUTPUT_DIR_PATH_WIN
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    ``: "{"..."}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"
    """)
    # Colorama initialization
    colorama_init()

    # Get columns
    columns = {}
    columnsByPortion = {portionName: {} for portionName in portionsOutputDirPath.keys()}
    for portionName, portionPath in portionsOutputDirPath.items():
        content_paths = [Path(dirObj) for dirObj in os.scandir(portionPath)]
        content_names = "\n  ".join(sorted([path.name for path in content_paths]))
        dirRelativePath = portionPath.absolute().relative_to(rootDirectory)
        logger.info(f"""Reading files from the directory "{dirRelativePath}". Below are its contents:""")
        for fpath in sorted(content_paths):
            logger.info(f"""  {fpath.name}""")
        for file in content_paths:
            conditions = PORTION_FILE_CRITERIA_DICT[portionName]
            conditionResults = [func(file) for func in conditions]
            if all(conditionResults):
                logger.debug(f"""  Reading "{file.absolute().relative_to(rootDirectory)}".""")
                df = pd.read_csv(file, dtype=str, nrows=10)
                columns[file.name] = df.columns
                columnsByPortion[portionName][file.name] = df.columns

    # Get all columns by file
    logger.info("""Printing columns by file.""")
    allColumns = set()
    it = 0
    columnsOrdered = OrderedDict(sorted(columns.items()))
    for key, value1 in columnsOrdered.items():
        if it > -1:
            logger.info(key)
            logger.info("")
            for el in sorted(value1):
                logger.info(f"  {el}")
                allColumns.add(el)
            logger.info("")
        it += 1

    # Get all columns by portion and file
    logger.info("""Printing columns by portion and file.""")
    allColumnsByPortion = OrderedDict({portionName: set() for portionName in sorted(columnsByPortion.keys())})
    columnsByPortionOrdered = OrderedDict(sorted(columnsByPortion.items()))
    for portionName, di in columnsByPortionOrdered.items():
        logger.info(f"{portionName}")
        for fileName, value2 in di.items():
            logger.info(f"  {fileName}")
            for el in sorted(value2):
                logger.info(f"    {el}")
                allColumnsByPortion[portionName].add(el)
            logger.info("")

    # Print the set of all columns
    logger.info("""Printing the set of all columns.""")
    for el in sorted(list(allColumns)):
        logger.info(f"  {el}")
    logger.info("")

    # Print the set of all columns by portion
    logger.info("""Print set of columns by portion.""")
    for portionName, columnsSet in allColumnsByPortion.items():
        logger.info(f"""{portionName}""")
        for columnName in sorted(list(columnsSet)):
            if columnName in COMPARISON_SET:
                text = f"""{Fore.LIGHTCYAN_EX}{columnName}{Style.RESET_ALL}"""
            else:
                text = columnName
            logger.info(f"  {text}")
        logger.info("")

    # End script
    logger.info(f"""Finished running "{thisFilePath.relative_to(projectDir)}".""")



"""
"""

import logging
import re
from pathlib import Path
from typing import List
from typing_extensions import (Union,
                               Tuple)
# Third-party packages
import pandas as pd
# First-party packages
from drapi.code.drapi.drapi import readDataFile


def expandColumn(tableOrPath: Union[pd.DataFrame, Path, str],
                 columnToSplit: Union[int, str],
                 nameOfNewColumns: List[Union[int, str]],
                 locationOfNewColumns: List[int],
                 splittingPattern: str,
                 logger: logging.Logger) -> pd.DataFrame:
    """
    """
    _ = logger

    if isinstance(tableOrPath, (str)):
        fpath = Path(tableOrPath)
        table = readDataFile(fname=fpath)
        table = pd.DataFrame(table)  # For type hinting
    elif isinstance(tableOrPath, Path):
        table = readDataFile(fname=tableOrPath)
        table = pd.DataFrame(table)  # For type hinting
    else:
        table = tableOrPath

    def splittingFunction(string_: str) -> Tuple[str]:
        """
        """
        reObj = re.search(splittingPattern, string_)
        if reObj:
            return reObj.groups()
        else:
            return None

    series0 = table[columnToSplit].apply(splittingFunction)

    table = table.drop(columns=columnToSplit)
    for it, (columnName, location) in enumerate(zip(nameOfNewColumns, locationOfNewColumns)):
        series = series0.apply(lambda tu: tu[it])
        table.insert(loc=location,
                     column=columnName,
                     value=series)

    return table


def expandColumnWrapper(runOutputPath: Path,
                        tableOrPath: Union[pd.DataFrame, Path, str],
                        columnToSplit: Union[int, str],
                        nameOfNewColumns: List[Union[int, str]],
                        locationOfNewColumns: List[int],
                        splittingPattern: str,
                        logger: logging.Logger,
                        separator: str = ",") -> None:
    """
    """
    newTable = expandColumn(tableOrPath=tableOrPath,
                            columnToSplit=columnToSplit,
                            nameOfNewColumns=nameOfNewColumns,
                            locationOfNewColumns=locationOfNewColumns,
                            splittingPattern=splittingPattern,
                            logger=logger)
    newTable.to_csv(runOutputPath,
                    index=False,
                    sep=separator)



"""
Makes de-identification maps, building on existing maps.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects all files to be CSV files. This is because it uses "pd.read_csv".
"""

import re
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.drapi import getPercentDifference, getTimestamp, makeDirPath


def concatenateMaps(NEW_MAPS_DIR_PATH,
                    OLD_MAPS_DIR_PATH,
                    pipelineOutputDir,
                    logger,
                    ROOT_DIRECTORY,
                    rootDirectory):
    """
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    makeDirPath(runOutputDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # Map new maps to variable names
    logger.info("""Mapping new maps to variable names.""")
    pattern = r"^([a-zA-Z_0-9\(\) ]+) map"
    newMapsFileDict = {}
    for fpath in NEW_MAPS_DIR_PATH.iterdir():
        if fpath.is_file():
            fname = fpath.stem
            obj = re.match(pattern, fname)
            if obj:
                variableName = obj.groups()[0]
            else:
                raise
            newMapsFileDict[variableName] = [fpath]
        else:
            pass

    # Match old and new maps by variable name, accounting for aliases.
    logger.info("""Matching old and new maps by variable name.""")
    variableNameSet = set()
    variableNameSet.update(OLD_MAPS_DIR_PATH.keys())
    variableNameSet.update(newMapsFileDict.keys())
    matchedMaps = {variableName: [] for variableName in sorted(list(variableNameSet))}
    for variableName, li in OLD_MAPS_DIR_PATH.items():
        matchedMaps[variableName].extend(li)
    for variableName, li in newMapsFileDict.items():
        matchedMaps[variableName].extend(li)

    # Load, concatenate, and save maps by variable names
    logger.info("""Loading, concatenating, and saving maps by variable names.""")
    concatenatedMapsDict = {}
    for variableName, li in matchedMaps.items():
        logger.info(f"""  Working on variable "{variableName}".""")
        concatenatedMap = pd.DataFrame()
        for fpath in li:
            fpath = Path(fpath)
            logger.info(f"""    Working on map located at "{fpath.absolute().relative_to(rootDirectory)}".""")
            df = pd.read_csv(fpath)
            columns = df.columns[:-1].to_list()
            columns = columns + [f"deid_{variableName}_id"]  # NOTE: Hack. Conform de-identified column name to this format.
            df.columns = columns
            concatenatedMap = pd.concat([concatenatedMap, df])
        concatenatedMapsDict[variableName] = concatenatedMap
        fpath = runOutputDir.joinpath(f"{variableName} map.csv")
        concatenatedMap.to_csv(fpath, index=False)

    # Quality control
    results = {}
    for variableName, df in concatenatedMapsDict.items():
        uniqueIDs = len(df.iloc[:, 0].unique())
        numIDs = len(df)
        percentDifference = getPercentDifference(uniqueIDs, numIDs)
        results[variableName] = {"Unique IDs": uniqueIDs,
                                 "Total IDs": numIDs,
                                 "Percent Similarity": percentDifference}
    resultsdf = pd.DataFrame.from_dict(results, orient="index")
    logger.info(f"Concatenation summary:\n{resultsdf.to_string()}")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")

    return runOutputDir



"""
Shifts the dates in a data set according to the date shift map for a patient.
"""

import logging
import os
from pathlib import Path
# Third-party packages
import numpy as np
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)

# Arguments
PATIENT_DE_IDENTIFICATION_MAP_PATH = Path(r"..\Concatenated Results\data\output\concatenateMaps\2023-12-13 11-59-04\Patient Key map.csv")

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    ``: "{"..."}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Load patient de-identification map
    patientDeIdentificationMap = pd.read_csv(PATIENT_DE_IDENTIFICATION_MAP_PATH)

    # Assign date shift values
    array = np.random.randint(low=-30,
                              high=30,
                              size=len(patientDeIdentificationMap))
    patientDeIdentificationMap["Date Shift"] = array

    # Save results
    savepath = runOutputDir.joinpath("Date Shift Map.CSV")
    patientDeIdentificationMap.to_csv(savepath, index=False)

    # End script
    logger.info(f"""Finished running "{thisFilePath.relative_to(projectDir)}".""")



"""
Create a working directory from a template. The general structure of the directory to be made is as follows:

New Directory
├── code
├── data
│   ├── input
│   └── output
├── logs
├── sql
├── .gitignore
└── README.md

where "New Directory" is the name of the directory to be created
"""

import argparse
import json
import os
import sys
import shutil
from pathlib import Path
# First-party imports
from drapi.templates import PATH as templatesPath


def win2nixPath(string: str) -> str:
    """
    Converts a Windows-like path to a Linux-like path by changing the path separators.
    """
    platform = sys.platform
    platformSeparator = os.sep
    if platform == "darwin":
        newString = platformSeparator.join(string.split("\\"))
    elif platform == "linux":
        newString = platformSeparator.join(string.split("\\"))
    elif platform == "win32":
        newString = string
    else:
        raise Exception(f"""Unsupported operating system: "{platform}".""")
    return newString


ROOT_PATH = templatesPath.joinpath("Portions Templates").__str__()
# NOTE `optionsDict` expects `"path"` values to be in Windows format because it's later used by `win2nixPath`
optionsDict = {"BO": {"number": 2,
                      "path": r"\Multi-Portion Template\Intermediate Results\BO Portion Template"},
               "De-identification Suite": {"number": 1,
                                           "path": r"\Multi-Portion Template\Concatenated Results"},
               "General Script": {"number": 3,
                                  "path": r"\Multi-Portion Template\Intermediate Results\General Script Template"},
               "i2b2": {"number": 4,
                        "path": r"\Multi-Portion Template\Intermediate Results\i2b2 Portion Template"},
               "Multi-Portion Template": {"number": 0,
                                          "path": r"\Multi-Portion Template"},
               "Clinical Text": {"number": 5,
                         "path": r"\Multi-Portion Template\Intermediate Results\Clinical Text Portion Template"},
               "OMOP": {"number": 6,
                        "path": r"\Multi-Portion Template\Intermediate Results\OMOP Portion Template"}}

optionsDict2 = {values["number"]: {"name": name,
                                   "path": ROOT_PATH + win2nixPath(values["path"])} for name, values in optionsDict.items()}

optionsNumbers = {name: value for name, values in optionsDict.items() for key, value in values.items() if key == "number"}


def copyTemplateDirectory(templateChoice: int,
                          destinationPath: str) -> None:
    """
    Given a template selection `templateChoice`, copies the corresponding template directory to the destination path, `destinationPath`.
    """

    templateDirPath = Path(optionsDict2[templateChoice]["path"])

    shutil.copytree(src=templateDirPath,
                    dst=destinationPath)

    # Prepare template for use
    for fpath in Path(destinationPath).glob("./**/*.*"):
        # Remove placeholder files
        if fpath.name.lower() == ".deleteme":
            os.remove(fpath)
        # Modify ".gitignore"
        if fpath.name.lower() == ".gitignore":
            with open(fpath, "r") as file:
                text = file.read()
                text = text.replace("!**/launchIPython.bat\n", "")
                text = text.replace("!.env", ".env")
                text = text.replace("!data/input/\n", "")
                text = text.replace("!data/output/\n", "")
            with open(fpath, "w") as file:
                file.write(text)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("templateChoice", help=f"""The template you wish to copy. Each template has a numerical option: {json.dumps(optionsNumbers)}""", choices=sorted(list(optionsDict2.keys())), type=int)

    parser.add_argument("destinationPath", help="The directory path for `templateChoice`.", type=str)

    args = parser.parse_args()

    templateChoice = args.templateChoice
    destinationPath = args.destinationPath

    try:
        copyTemplateDirectory(templateChoice=templateChoice,
                              destinationPath=destinationPath)
    except Exception as err:
        shutil.rmtree(destinationPath)
        raise err



"""
OMOP Data Pull Script

Please be sure to review the README for this script before running.
"""

__all__ = []

import glob
import logging
import os
import sys
from pathlib import Path
# Third-party libraries
import pandas as pd
import sqlalchemy as sa
import yaml
# Super-local libraries
import deidentify
from drapi.code.drapi.drapi import (makeDirPath,
                                    getTimestamp)
from drapi.code.drapi.omop.configProcessing import (editConfig,
                                                    interpretPath)

# Arguments
LOG_LEVEL = "INFO"  # Lowest level available is "9"

# Arguments: SQL server settings
if False:
    # TODO Not implemented. See `db_connect`
    USERNAME_ENVIRONMENT_VARIABLE = "USER"
    PASSWORD_ENVIRONMENT_VARIABLE = "HFA_UFADPWD"
    SERVER = "DWSRSRCH01.shands.ufl.edu"  # AKA `HOST`
    DATABASE = "DWS_OMOP_PROD"
    USERDOMAIN = "UFAD"

# Variables
timestamp = getTimestamp()
config_file0 = os.path.join("Config1.YAML")
config_file = os.path.join("Data", "Output", "Config2.YAML")
editConfig(config_file0, config_file, timestamp)
base_dir = Path(__file__).parent.parent

# Variables: SQL server settings
if False:
    # TODO Not implemented. See `db_connect`
    username = os.environ[USERNAME_ENVIRONMENT_VARIABLE]
    userid = fr"{USERDOMAIN}\{username}"
    userpwd = os.environ[PASSWORD_ENVIRONMENT_VARIABLE]


def db_connect(host, database):
    # SQL Server settings
    SERVER = "DWSRSRCH01.shands.ufl.edu"  # AKA `HOST`
    DATABASE = "DWS_OMOP_PROD"
    USERDOMAIN = "UFAD"
    USERNAME = os.environ["USER"]
    UID = fr"{USERDOMAIN}\{USERNAME}"
    PWD = os.environ["HFA_UFADPWD"]
    connstr = f"mssql+pymssql://{UID}:{PWD}@{SERVER}/{DATABASE}"
    engine = sa.create_engine(connstr)
    return engine


def db_close(db_conn):
    db_conn.dispose()


def import_config():
    try:
        config_setting = yaml.safe_load(open(config_file))
    except Exception as e:
        logging.critical(e)
        sys.exit(1)
    return config_setting


def db_query(query, db_connection):
    dfs = pd.DataFrame()
    for chunk in pd.read_sql(query, db_connection, chunksize=50000):
        df = pd.DataFrame(chunk)
        dfs = pd.concat([dfs, df], ignore_index=False)
    return dfs


def db_execute_read_query(host, database, query):
    db_conn = db_connect(host, database)
    data = db_query(query, db_conn)
    db_close(db_conn)
    return data


def db_info(config):
    host = config['db_connections']['data_pull']['server']
    database = config['db_connections']['data_pull']['database']
    schema = config['db_connections']['data_pull']['schema']
    dict_of_dates = config['clinical_data_tables']
    dict_of_search = config['data_selection']
    person_id_file_path = interpretPath(search_config['person_id'])
    start_date = str(search_config['date_range']['start_date'])
    end_date = str(search_config['date_range']['end_date'])
    start_date = "'" + start_date + "'"
    end_date = "'" + end_date + "'"
    try:
        list_of_tables = list(dict_of_dates)
    except TypeError as error:
        message = error.args[0]
        if "object is not iterable" in message:
            list_of_tables = []
        else:
            raise
    return host, database, schema, list_of_tables, dict_of_dates, person_id_file_path, start_date, end_date, search_config, dict_of_search


def person_info_query(search_config):
    person_id_file_path = interpretPath(search_config['person_id'])
    list_of_tables = search_config['person_information_tables']['person_and_death']
    if (list_of_tables):
        for i in range(len(list_of_tables)):
            h = True
            m = 'w'
            current_table = list_of_tables[i]
            columns = search_config.get(current_table)
            columns_string = ', '.join(map(str, columns))
            for person_id in pd.read_csv(person_id_file_path, chunksize=500):
                person_id_list = person_id.iloc[:, 0]
                person_id_list_string = ', '.join(map(str, person_id_list))
                query_string = "SELECT " + columns_string + " FROM " + schema + "." + current_table + " WHERE person_id IN (" + person_id_list_string + ")"
                data = db_execute_read_query(host, database, query_string)
                if not data.empty:
                    identified_file_location_AsString = interpretPath(search_config['data_output']['identified_file_location'])
                    file_location = identified_file_location_AsString + current_table + '.csv'
                    csv_output_file = (file_location)
                    try:
                        data.to_csv(csv_output_file, index=False, header=h, mode=m)
                        h = False
                        m = 'a'
                    except Exception as err:
                        errorMessage = f"{err.__class__.__name__}: {err.__str__()}"
                        message = f"An exception has occurred. This may be caused by an output directory error; please double check the directory in the config file. The actual exception message is below:\n{errorMessage}"
                        logging.error(message)
    list_of_tables = search_config['person_information_tables']['location_table']
    if (list_of_tables and ('person' in search_config['person_information_tables']['person_and_death'])):
        identified_file_location_AsString = interpretPath(search_config['data_output']['identified_file_location'])
        person_file_path = identified_file_location_AsString + 'person.csv'
        h = True
        m = 'w'
        for location_id in pd.read_csv(person_file_path, chunksize=500):
            location_id_list = []
            location_id_list = location_id['location_id']
            location_id_list_string = ', '.join(map(str, location_id_list))
            current_table = list_of_tables[0]
            columns = search_config.get(current_table)
            columns_string = ', '.join(map(str, columns))
            query_string = "SELECT " + columns_string + " FROM " + current_table + " WHERE location_id IN (" + location_id_list_string + ")"
            data = db_execute_read_query(host, database, query_string)
            if not data.empty:
                identified_file_location_AsString = interpretPath(search_config['data_output']['identified_file_location'])
                file_location = identified_file_location_AsString + current_table + '.csv'
                csv_output_file = (file_location)
                try:
                    data.to_csv(csv_output_file, index=False, header=h, mode=m)
                    h = False
                    m = 'a'
                except Exception as err:
                    errorMessage = f"{err.__class__.__name__}: {err.__str__()}"
                    message = f"An exception has occurred. This may be caused by an output directory error; please double check the directory in the config file. The actual exception message is below:\n{errorMessage}"
                    logging.error(message)


def query_attempt(search_config, list_of_tables, dict_of_dates, person_id_file_path, start_date, end_date, dict_of_search):
    numTables = len(list_of_tables)
    for i in range(numTables):
        h = True
        m = 'w'
        h1 = True
        m1 = 'w'
        current_table = list_of_tables[i]
        message = f"""  Working on table "{current_table}", table {i+1} of {numTables}."""
        logging.info(message)
        columns = search_config.get(current_table)
        columns_string = ', '.join(map(str, columns))
        # Get number of chunks of table
        numChunks = sum([1 for _ in pd.read_csv(person_id_file_path, chunksize=500)])
        for it, person_id in enumerate(pd.read_csv(person_id_file_path, chunksize=500), start=1):
            logging.info(f"""    Working on table chunk {it} of {numChunks}""")
            person_id_list = person_id.iloc[:, 0]
            person_id_list_string = ', '.join(map(str, person_id_list))
            date_sorted_by = dict_of_dates.get(current_table)
            query_string = "(SELECT " + columns_string + " FROM " + current_table + " WHERE person_id IN (" + person_id_list_string + ") AND " + date_sorted_by + " BETWEEN " + start_date + " AND " + end_date + ")"
            if ((dict_of_search.get(current_table) is not None)):
                row_query_string = ''
                rows = dict_of_search.get(current_table)
                before, sep, after = current_table.partition('_')
                current_key = before
                if (current_table == 'condition_occurrence'):
                    row_query_string = "("
                    for i in range(len(rows)):
                        row_string = rows[i]
                        row_query_string += "(SELECT " + columns_string + " FROM " + current_table + " WHERE " + current_key + "_source_value LIKE '" + row_string + "%' AND " + date_sorted_by + " BETWEEN " + start_date + " AND " + end_date + ")"
                        if (i != len(rows) - 1):
                            row_query_string += " UNION "
                        else:
                            row_query_string += ")"
                elif (current_table == 'measurement'):
                    rows_string = "', '".join(map(str, rows))
                    row_query_string = "(SELECT " + columns_string + " FROM " + current_table + " WHERE " + current_key + "_source_value IN ('" + rows_string + "') AND " + date_sorted_by + " BETWEEN " + start_date + " AND " + end_date + ")"
                    if (dict_of_search.get('measurement_laboratory_search') is not None):
                        row_string = "', '".join(map(str, dict_of_search.get('measurement_laboratory_search')))
                        search_query_string = "(SELECT " + columns_string + " FROM " + current_key + " WHERE " + current_key + "_source_value IN ('" + row_string + "') AND " + date_sorted_by + " BETWEEN " + start_date + " AND " + end_date + ")"
                    else:
                        search_query_string = "(SELECT " + columns_string + " FROM " + current_key + " WHERE " + current_key + "_source_value LIKE '%[0-9]-[0-9]'  AND " + date_sorted_by + " BETWEEN " + start_date + " AND " + end_date + ")"
                    query = query_string + " INTERSECT " + search_query_string + " ORDER BY person_id, " + date_sorted_by
                    logging.log(9, f"  ..  >>> This is the query >>>\n{query}\n<<< End of query <<<")
                    data1 = db_execute_read_query(host, database, query)
                    identified_file_location_asString = interpretPath(search_config['data_output']['identified_file_location'])
                    file_location = identified_file_location_asString + 'measurement_laboratories.csv'
                    if not data1.empty:
                        csv_output_file = (file_location)
                        try:
                            logging.log(9, f"""  ..  Saving query chunk results to "{csv_output_file}".""")
                            data1.to_csv(csv_output_file, index=False, header=h1, mode=m1)
                            h1 = False
                            m1 = 'a'
                        except Exception as err:
                            errorMessage = f"{err.__class__.__name__}: {err.__str__()}"
                            message = f" . . . . . . .An exception has occurred. This may be caused by an output directory error; please double check the directory in the config file. The actual exception message is below:\n{errorMessage}"
                            logging.error(message)
                else:
                    rows_string = "', '".join(map(str, rows))
                    row_query_string = "(SELECT " + columns_string + " FROM " + current_table + " WHERE " + current_key + "_source_value IN ('" + rows_string + "') AND " + date_sorted_by + " BETWEEN " + start_date + " AND " + end_date + ")"
                query = query_string + " INTERSECT " + row_query_string
                query_string = query
            query_string += " ORDER BY person_id, " + date_sorted_by
            logging.log(9, f"  ..  >>> This is the query >>>\n{query_string}\n<<< End of query <<<")
            data = db_execute_read_query(host, database, query_string)
            if not data.empty:
                identified_file_location_asString = interpretPath(search_config['data_output']['identified_file_location'])
                file_location = identified_file_location_asString + current_table + '.csv'
                csv_output_file = (file_location)
                try:
                    logging.info(f"""  ..  Saving query chunk results to "{csv_output_file}".""")
                    data.to_csv(csv_output_file, index=False, header=h, mode=m)
                    h = False
                    m = 'a'
                except Exception as err:
                    errorMessage = f"{err.__class__.__name__}: {err.__str__()}"
                    message = f"An exception has occurred. This may be caused by an output directory error; please double check the directory in the config file. The actual exception message is below:\n{errorMessage}"
                    logging.error(message)
        logging.info(f"""  Done processing table "{current_table}".""")


def deidentify_(mapping_file_location):
    data_release = search_config.get('data_release')
    if ('deidentified' in data_release or 'limited' in data_release):
        path = interpretPath(search_config['data_output']['identified_file_location'])
        map_path = interpretPath(search_config['data_output']['mapping_location'])
        all_files = glob.glob(path + "/*.csv")
        map_files = glob.glob(map_path + "/*.csv")
        for file in all_files:
            current_table = os.path.basename(file)
            file_location = search_config['data_output']['deidentified_file_location'] + data_release[0] + "-" + current_table
            deidentified_file = deidentify.shift_person_occurrence(file, map_files, data_release)
            deidentified_file.to_csv(file_location, index=False)


if __name__ == '__main__':
    # Logging block
    loglevel = "DEBUG"
    this_file_path = Path(__file__)
    logpath = os.path.join(base_dir, "logs", f"log {timestamp}.log")
    makeDirPath(Path(logpath).parent)
    fileHandler = logging.FileHandler(logpath)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(loglevel)
    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler,
                                  streamHandler],
                        level=loglevel)

    logging.info("Starting program.")
    # import search configuration
    search_config = import_config()
    # import connection information for DB connection
    host, database, schema, list_of_tables, dict_of_dates, person_id_file_path, start_date, end_date, search_config, dict_of_search = db_info(search_config)

    logging.info(f"""Reading cohort from `person_id_file_path`: "{person_id_file_path}".""")

    for dataType, path in search_config["data_output"].items():
        path = Path(interpretPath(path))
        makeDirPath(path)

    if (list_of_tables):
        query_attempt(search_config, list_of_tables, dict_of_dates, person_id_file_path, start_date, end_date, dict_of_search)
    if "person_information_tables" in search_config.keys():
        person_info_query(search_config)
    # deidentify_(mapping_file_location)
    logging.info("Finished program.")



"""
Statistics functions
"""

import pandas as pd


def fiveNumberSummary(series: pd.Series, more=True) -> pd.Series:
    """
    """
    x1 = series.min()
    x2 = series.quantile(.25)
    x3 = series.quantile(.50)
    y3 = series.mean()
    x4 = series.quantile(.75)
    x5 = series.quantile(.90)
    x6 = series.quantile(.95)
    x7 = series.quantile(.99)
    x8 = series.quantile(.999)
    x9 = series.max()
    sumStatsValues1 = [x1, x2, x3, x4, x5, x6, x7, x8, x9]
    sumStatsValues2 = ["", "", y3, "", "", "", "", "", ""]
    fiveNumberSummaryColumns = ["min",
                                "25th ptile",
                                "median",
                                "75th ptile",
                                "max"]
    sumStatsNames1 = ["min",
                      "25th ptile",
                      "median",
                      "75th ptile",
                      "90th ptile",
                      "95th ptile",
                      "99th ptile",
                      "99.9th ptile",
                      "max"]
    sumStatsNames2 = ["",
                      "",
                      "mean",
                      "",
                      "",
                      "",
                      "",
                      "",
                      ""]
    summaryStats = pd.DataFrame([sumStatsNames1, sumStatsValues1, sumStatsValues2, sumStatsNames2]).T
    summaryStats.columns = ["1", "Percentiles", "Means", "2"]
    summaryStats.index = range(1, len(summaryStats) + 1)
    summaryStats[["Percentiles", "Means"]] = summaryStats[["Percentiles", "Means"]].applymap(lambda el: "" if isinstance(el, str) else f"{int(el):,}")

    if more:
        pass
    else:
        summaryStats = summaryStats.set_index("1")
        colsToDrop = [col for col in summaryStats.T.columns if col not in fiveNumberSummaryColumns]
        summaryStats = summaryStats.drop(labels=colsToDrop)
        summaryStats = summaryStats.reset_index()

    summaryStats = summaryStats.rename(columns={"1": "", "2": ""})

    return summaryStats



"""
Download clinical notes using DRAPI.

This is a template script which should be moved to `makeDirTempalte` after it's implemented.
"""

import logging
import os
from pathlib import Path
# Third-party packages
from sqlalchemy import create_engine
# Local packages
from drapi.drapi import getTimestamp, makeDirPath, successiveParents
from drapi.notes.freeText import main

# Arguments
_ = None

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 1
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"
connection = create_engine(conStr).connect().execution_options(stream_results=True)

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
# logFormat = logging.Formatter(f"""[%(asctime)s]\n[%(levelname)s](%(funcName)s)\n{"-"*24}> %(message)s""")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments
    ``: "{""}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Pipeline arguments
    COHORT_NAME = ""                 # An arbitrary name used in file names
    COHORT_FILE = ""                 # A file name that is located directory specified by the variable `data_dir`
    IRB_NUMBER = ""                  # Used for creating the de-identification map IDs.
    ID_TYPE = ""                     # Pick from "EncounterCSN", "EncounterKey", or "PatientKey". Choose the ID type you used in `COHORT_FILE`
    NOTE_VERSION = ""                # Pick from "all", or "last"
    DE_IDENTIFICATION_MODE = ""      # Pick from "deid", "lds", or "phi"
    LOG_LEVEL = ""                   # See the "logging" module for valid values for the `loglevel` parameter.
    SQL_ENCOUNTER_EFFECTIVE_DATE_START = '2012-01-01'  # The beginning of date range of encounters to collect. Format: YYYY-MM-DD
    SQL_ENCOUNTER_EFFECTIVE_DATE_END = '2023-12-31'  # The end of date range of encounters to collect. Format: YYYY-MM-DD

    # Arguments: SQL connection settings
    USERDOMAIN = "UFAD"
    USERNAME = os.environ["USER"]
    UID = fr"{USERDOMAIN}\{USERNAME}"
    PWD = os.environ["HFA_UFADPWD"]

    main()



"""
SQL helper functions
"""

import os
from pathlib import Path
# Third-party libraries
pass
# Local libraries
pass

# Arguments
MODULE_ROOT_DIRECTORY_PATH = Path(__file__).absolute().parent.parent.parent

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"


def labelMRNColumns(query):
    """
    NOTE Not implemented

    Replace text in a query for each of the MRN columsn thusly:

    `Table__1308.IDENT_ID_INT` --> `Table__1308.IDENT_ID_INT as MRN_UF`
    `Table__1311.IDENT_ID_int` --> `Table__1311.IDENT_ID_int as MRN_Vista`
    `Table__1312.IDENT_ID_int` --> `Table__1312.IDENT_ID_int as MRN_Rehab`
    `Table__1117.IDENT_ID_INT` --> `Table__1117.IDENT_ID_INT as MRN_Jax`
    `Table__2699.IDENT_ID_INT` --> `Table__2699.IDENT_ID_INT as MRN_CFH_Lessburg`
    `Table__2700.IDENT_ID_INT` --> `Table__2700.IDENT_ID_INT AS MRN_CFH_Villages`

    Outline

    for each instance of the above strings, if it's not behind a comment marker (e.g., `--`), replace it with its corresponding string pair. If it's the last column make sure to no add a comma
    """
    # TODO
    return



"""
Convert person ID column to patient key

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects all files to be CSV files. This is because it uses "pd.read_csv".
"""

from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.idealist.idealist import idealistMap2dict
from drapi.drapi import getTimestamp, makeDirPath


def convertColumns(COLUMNS_TO_CONVERT_DI,
                   PERSON_ID_MAP_PATH: Path,
                   listOfPortionDirs,
                   LIST_OF_PORTION_CONDITIONS,
                   CHUNK_SIZE,
                   pipelineOutputDir,
                   logger,
                   ROOT_DIRECTORY,
                   rootDirectory) -> Path:
    """
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    makeDirPath(runOutputDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # Load person ID map
    personIDMap = pd.read_csv(PERSON_ID_MAP_PATH)
    personIDMapDi = idealistMap2dict(personIDMap, "person_id", "patient_key")

    # Convert columns
    logger.info("""Getting the set of values for each variable to convert.""")
    # Create file parameters dictionary

    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logger.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        for file in directory.iterdir():
            logger.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logger.info("""    File has met all conditions for processing.""")
                numChunks = sum([1 for _ in pd.read_csv(file, chunksize=CHUNK_SIZE)])
                dfChunks = pd.read_csv(file, chunksize=CHUNK_SIZE)
                for it, dfChunk in enumerate(dfChunks, start=1):
                    dfChunk = pd.DataFrame(dfChunk)
                    logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                    for columnName in dfChunk.columns:
                        logger.info(f"""  ..    Working on column "{columnName}".""")
                        if columnName in COLUMNS_TO_CONVERT_DI.keys():
                            logger.info("""  ..  ..  Column must be converted. Converting values.""")
                            dfChunk[columnName] = dfChunk[columnName].apply(lambda IDNum: personIDMapDi[IDNum])
                            dfChunk = dfChunk.rename(columns={columnName: COLUMNS_TO_CONVERT_DI[columnName]})
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    logger.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")
            else:
                logger.info("""    This file does not need to be processed.""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")

    return runOutputDir



"""
The OneFlorida workflow.
"""

ID_TYPE_LIST = ["MRN (Jax)",
                "MRN (Pathology)",
                "MRN (UF)",
                "OneFlorida Patient ID",
                "Patient Key"]
ID_TYPE_DICT = {"oneflorida patient id": "A.PATID",
                "patient key": "A.PATNT_KEY",
                "mrn (uf)": "B.IDENT_ID",
                "mrn (jax)": "C.IDENT_ID",
                "mrn (pathology)": "D.IDENT_ID"}



"""
Convert columns.

# NOTE This is a more general form of the OMOP-only version, "convertColumns.py"
# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# TODO Assign portion name to each path (per OS) so that portion files are stored in their respective folders, this prevents file from being overwritten in the unlikely, but possible, case files from different portions have the same name.
"""

import logging
import os
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
import sqlite3
# Local packages
from drapi.code.drapi.drapi import (flatExtend,
                                    getTimestamp,
                                    makeDirPath,
                                    makeMap,
                                    readDataFile,
                                    successiveParents)
# Local packages: Script parameters: General
from common import (IRB_NUMBER,
                    ALIAS_DATA_TYPES,
                    DATA_REQUEST_ROOT_DIRECTORY_DEPTH,
                    DATA_TYPES_DICT,
                    VARIABLE_ALIASES,
                    VARIABLE_SUFFIXES)
# Local packages: Script parameters: Paths
from common import (MODIFIED_OMOP_PORTION_DIR_MAC, MODIFIED_OMOP_PORTION_DIR_WIN,
                    NOTES_PORTION_DIR_MAC, NOTES_PORTION_DIR_WIN,
                    OMOP_PORTION_DIR_MAC, OMOP_PORTION_DIR_WIN,
                    OMOP_PORTION_2_DIR_MAC, OMOP_PORTION_2_DIR_WIN)
# Local packages: Script parameters: File criteria
from common import (NOTES_PORTION_FILE_CRITERIA,
                    OMOP_PORTION_FILE_CRITERIA)

# Arguments
if True:
    MAPS = {"provider_id": {"to": "Provider Key",
                            "mapPath": r"../Concatenated Results/data/output/mapToSqlite/2024-02-23 17-36-14/provider_id to Provider Key.db"}}
    MAC_PATHS = [Path("/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/Concatenated Results/data/output/convertColumnsGeneral/2024-01-24 17-48-25"),
                 Path("/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/Concatenated Results/data/output/convertColumnsGeneral/2024-01-26 20-17-44")]
    WIN_PATHS = []
    LIST_OF_PORTION_CONDITIONS = [OMOP_PORTION_FILE_CRITERIA,
                                  OMOP_PORTION_FILE_CRITERIA]
elif False:
    MAPS = {"person_id": {"to": "PatientKey",
                          "mapPath": r"..\Concatenated Results\data\output\mapToSqlite\2024-01-22 20-58-28\person_id to PatientKey.db"}}
    MAC_PATHS = [OMOP_PORTION_DIR_MAC]
    WIN_PATHS = [OMOP_PORTION_DIR_WIN]
    LIST_OF_PORTION_CONDITIONS = [OMOP_PORTION_FILE_CRITERIA + [lambda pathObj: pathObj.stem != "condition_occurrence"]]
elif True:
    MAPS = {"person_id": {"to": "PatientKey",
                          "mapPath": r"..\Concatenated Results\data\output\mapToSqlite\2024-01-22 20-58-28\person_id to PatientKey.db"},
            "visit_occurrence_id": {"to": "Encounter # (CSN)",
                                    "mapPath": r"..\Concatenated Results\data\output\mapToSqlite\2024-01-22 21-11-13\visit_occurrence_id to Encounter # (CSN).db"}}
    MAC_PATHS = [OMOP_PORTION_DIR_MAC,
                 OMOP_PORTION_2_DIR_MAC]
    WIN_PATHS = [OMOP_PORTION_DIR_WIN,
                 OMOP_PORTION_2_DIR_WIN]
    previouslyDoneFiles = flatExtend([[fpath.stem for fpath in Path(r"..\Concatenated Results\data\output\convertColumnsGeneral\2024-01-24 17-48-25").iterdir()],
                                      []])
    newOMOPFileCriteria = OMOP_PORTION_FILE_CRITERIA + [lambda pathObj: pathObj.stem not in previouslyDoneFiles]
    LIST_OF_PORTION_CONDITIONS = [newOMOPFileCriteria,
                                  newOMOPFileCriteria]
elif True:
    MAPS = {"PatientKey": {"to": "De-identified PatientKey",
                           "mapPath": r""},
            "EncounterCSN": {"to": "De-identified EncounterCSN",
                             "mapPath": r""}}
    MAC_PATHS = [MODIFIED_OMOP_PORTION_DIR_MAC,
                 NOTES_PORTION_DIR_MAC]
    WIN_PATHS = [MODIFIED_OMOP_PORTION_DIR_WIN,
                 NOTES_PORTION_DIR_WIN]
    LIST_OF_PORTION_CONDITIONS = [OMOP_PORTION_FILE_CRITERIA,
                                  NOTES_PORTION_FILE_CRITERIA]

# Arguments; General
CHUNK_SIZE = 50000
MESSAGE_MODULO_CHUNKS = 100  # How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
# ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
#  "IRB_DIRECTORY",                   # noqa
#  "DATA_REQUEST_DIRECTORY",          # noqa
#  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "linux":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Functions

def choosePathToLog(path: Path, rootPath: Path) -> Path:
    """
    Decides if a path is a subpath of `rootPath`. If it is, display it reltaive to `rootPath`. If it is not, display it as an absolute path.
    """
    commonPath = os.path.commonpath([path.absolute(), rootPath.absolute()])

    lenCommonPath = len(commonPath)
    lenRootPath = len(str(rootPath.absolute()))
    if lenCommonPath < lenRootPath:
        pathToDisplay = path
    elif lenCommonPath >= lenRootPath:
        pathToDisplay = path.absolute().relative_to(rootPath)
    else:
        raise Exception("An unexpected error occurred.")
                                    
    return pathToDisplay


if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # Assert all portions have file criteria accounted for
    if len(listOfPortionDirs) == len(LIST_OF_PORTION_CONDITIONS):
        pass
    else:
        raise Exception("The number of portions does not equal the number of portion conditions.")

    # Collect columns to convert
    columnsToConvert = sorted(list(MAPS.keys()))

    # Load de-identification maps for each variable that needs to be de-identified
    logging.info("""Loading de-identification maps for each variable that needs to be de-identified.""")
    mapsDi = {}
    mapsColumnNames = {}
    variablesCollected = [fname for fname in MAPS.keys()]
    for varName in variablesCollected:
        logging.info(f"""  Loading map for "{varName}".""")
        toVariable = MAPS[varName]["to"]
        varPath = MAPS[varName]["mapPath"]
        tableName = f"{varName} to {toVariable}"
        with sqlite3.connect(varPath) as connection:
            map_ = pd.read_sql_query(sql=f"SELECT * FROM '{tableName}' LIMIT 1", con=connection)
        mapsDi[varName] = map_
        mapsColumnNames[varName] = map_.columns[-1]
    # Add aliases to `mapsColumnNames`
    for varName in VARIABLE_ALIASES.keys():
        varAlias = VARIABLE_ALIASES[varName]
        map_ = makeMap(IDset=set(),
                       IDName=varName,
                       startFrom=1,
                       irbNumber=IRB_NUMBER,
                       suffix=VARIABLE_SUFFIXES[varAlias]["deIdIDSuffix"],
                       columnSuffix=varName,
                       deIdentificationMapStyle="lemur",
                       logger=logging.getLogger())
        mapsColumnNames[varName] = map_.columns[-1]
    # Add aliases to `DATA_TYPES_DICT`
    DATA_TYPES_DICT.update(ALIAS_DATA_TYPES)

    # De-identify columns
    logging.info("""De-identifying files.""")
    allowLogging = False
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logging.info(f"""Working on directory "{choosePathToLog(directory, rootDirectory)}".""")
        for file in directory.iterdir():
            logging.info(f"""  Working on file "{choosePathToLog(file, rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logging.info("""    File has met all conditions for processing.""")
                logging.info("""  ..  Reading file to count the number of chunks.""")
                numChunks = sum([1 for _ in readDataFile(file, chunksize=CHUNK_SIZE, low_memory=False)])
                logging.info(f"""  ..  This file has {numChunks:,} chunks.""")
                # Calculate logging requency
                if numChunks < MESSAGE_MODULO_CHUNKS:
                    moduloChunks = numChunks
                else:
                    moduloChunks = round(numChunks / MESSAGE_MODULO_CHUNKS)
                if numChunks / moduloChunks < 100:
                    moduloChunks = 1
                else:
                    pass
                dfChunks = readDataFile(file, chunksize=CHUNK_SIZE, low_memory=False)
                filePreview = readDataFile(file, nrows=2)
                filePreview = pd.DataFrame(filePreview)
                nasumDict = {columnName: {"nasumRows": 0,
                                          "nasumValues": 0,
                                          "nasumRowsTotal": 0,
                                          "nasumValuesTotal": 0} for columnName in filePreview.columns}
                for it, dfChunk0 in enumerate(dfChunks, start=1):
                    dfChunk = pd.DataFrame(dfChunk0)
                    # Work on chunk
                    if it % moduloChunks == 0:
                        allowLogging = True
                    else:
                        allowLogging = False
                    if allowLogging:
                        logging.info(f"""  ..  Working on chunk {it:,} of {numChunks:,}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        if allowLogging:
                            logging.info(f"""  ..    Working on column "{columnName}".""")
                        if columnName in columnsToConvert:
                            if allowLogging:
                                logging.info("""  ..  ..  Column must be converted. Converting column...""")
                            # Get position of (old) column
                            oldColumnPosition = dfChunk.columns.get_loc(columnName)
                            # Build map for chunk
                            databasePath = MAPS[columnName]["mapPath"]
                            toVariable = MAPS[columnName]["to"]
                            tableName = f"{columnName} to {toVariable}"
                            with sqlite3.connect(database=databasePath) as connection:
                                cursor = connection.cursor()
                                uniqueValues = dfChunk[columnName].dropna().drop_duplicates()
                                queryFilterArgument = ",".join(uniqueValues.astype(str).to_list())
                                query = f"""SELECT
                                                A.'{columnName}'
                                                ,A.'{toVariable}'
                                            FROM
                                                '{tableName}' AS A
                                            WHERE
                                                A.'{columnName}' in ({queryFilterArgument})
                                         """
                                cursor.execute(query)
                                sqliteResult = cursor.fetchall()
                                chunkMapColumnNames = [columnName, toVariable]
                                chunkMap = pd.DataFrame(sqliteResult, columns=chunkMapColumnNames)
                            # Add new column
                            dfChunk = dfChunk.join(other=chunkMap.set_index(columnName),
                                                   on=columnName,
                                                   how="outer",
                                                   lsuffix="",
                                                   rsuffix="_R")
                            # Remove old column and extra column
                            _ = dfChunk.pop(columnName)
                            # Move new column to old column position
                            newColumnName = mapsColumnNames[columnName]
                            _ = dfChunk.insert(loc=oldColumnPosition,
                                               column=newColumnName,
                                               value=dfChunk.pop(newColumnName))
                            # QA: Count un-mapped rows and values. Update total count for file
                            columnSeries = pd.Series(dfChunk[newColumnName])
                            nasumRows = columnSeries.isna().sum()
                            nasumValues = len(uniqueValues) - uniqueValues.isin(chunkMap[columnName]).sum()
                            nasumDict[columnName]["nasumRows"] = nasumRows
                            nasumDict[columnName]["nasumValues"] = nasumValues
                            nasumDict[columnName]["nasumRowsTotal"] += nasumRows
                            nasumDict[columnName]["nasumValuesTotal"] += nasumValues
                            if nasumValues > 0:
                                logging.warning(f"""  ..  WARNING: there were {nasumRows:,} missing ID values (NaNs) for "{columnName}". This might be an indication of failure to map values. The number is not a count of unique values.""")
                            if nasumValues > 0:
                                logging.warning(f"""  ..  WARNING: There were {nasumValues:,} missing unique ID values (NaNs) for "{columnName}". This might be an indication of failure to map values.""")
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    if allowLogging:
                        logging.info(f"""  ..  Chunk saved to "{choosePathToLog(exportPath, rootDirectory)}".""")
                nasumRowsTotal = nasumDict[columnName]["nasumRowsTotal"]
                nasumValuesTotal = nasumDict[columnName]["nasumValuesTotal"]
                nasumdf = pd.DataFrame.from_dict(data=nasumDict)
                warningText = f"""  ..  WARNING: Below are the missing value summaries. Missing values are NaNs. Missing values may be an indication of failure to map values (as in the case of IDs).\n{nasumdf.to_string()}"""
                logging.warning(warningText)
            else:
                logging.info("""    This file does not need to be processed.""")

    # Output location summary
    logging.info(f"""Script output is located in the following directory: "{choosePathToLog(runOutputDir, rootDirectory)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")
